@@ -7,7 +7,9 @@
   |   See COPYING                                                           |
   +-------------------------------------------------------------------------+ */
 
+#include <mrpt/core/lock_helper.h>
 #include <mrpt/obs/CObservation3DRangeScan.h>
+#include <mrpt/obs/CObservationIMU.h>
 #include <mrpt/obs/CObservationPointCloud.h>
 #include <mrpt/system/CTicTac.h>
 #include <mrpt/system/filesystem.h>
@@ -22,6 +24,7 @@
 #include <geometry_msgs/PoseArray.h>
 #include <geometry_msgs/PoseWithCovarianceStamped.h>
 #include <mrpt/ros1bridge/image.h>
+#include <mrpt/ros1bridge/imu.h>
 #include <mrpt/ros1bridge/laser_scan.h>
 #include <mrpt/ros1bridge/map.h>
 #include <mrpt/ros1bridge/point_cloud2.h>
@@ -30,6 +33,7 @@
 #include <nav_msgs/MapMetaData.h>
 #include <nav_msgs/Odometry.h>
 #include <sensor_msgs/Image.h>
+#include <sensor_msgs/Imu.h>
 #include <sensor_msgs/LaserScan.h>
 #include <sensor_msgs/PointCloud2.h>
 #include <tf2_geometry_msgs/tf2_geometry_msgs.h>
@@ -42,6 +46,7 @@ using Msg_TransformStamped = geometry_msgs::TransformStamped;
 //                    ROS 2
 // ===========================================
 #include <mrpt/ros2bridge/image.h>
+#include <mrpt/ros2bridge/imu.h>
 #include <mrpt/ros2bridge/laser_scan.h>
 #include <mrpt/ros2bridge/map.h>
 #include <mrpt/ros2bridge/point_cloud2.h>
@@ -51,6 +56,7 @@ using Msg_TransformStamped = geometry_msgs::TransformStamped;
 #include <nav_msgs/msg/odometry.hpp>
 #include <nav_msgs/srv/get_map.hpp>
 #include <sensor_msgs/msg/image.hpp>
+#include <sensor_msgs/msg/imu.hpp>
 #include <sensor_msgs/msg/laser_scan.hpp>
 #include <sensor_msgs/msg/point_cloud2.hpp>
 
@@ -125,7 +131,15 @@ MVSimNode::MVSimNode(rclcpp::Node::SharedPtr& n)
 	do_fake_localization_ = n_->declare_parameter<bool>(
 		"do_fake_localization", do_fake_localization_);
 
+	publisher_history_len_ = n_->declare_parameter<int>(
+		"publisher_history_len", publisher_history_len_);
+
 	// n_->declare_parameter("use_sim_time"); // already declared error?
+	if (true == n_->get_parameter_or("use_sim_time", false))
+	{
+		THROW_EXCEPTION(
+			"At present, MVSIM can only work with use_sim_time=false");
+	}
 #endif
 
 	// Launch GUI thread:
@@ -174,11 +188,17 @@ MVSimNode::MVSimNode(rclcpp::Node::SharedPtr& n)
 	localn_.param(
 		"do_fake_localization", do_fake_localization_, do_fake_localization_);
 
-	// In case the user didn't set it:
-	n_.setParam("/use_sim_time", true);
+	// JLBC: At present, mvsim does not use sim_time for neither ROS 1 nor
+	// ROS 2.
+	// n_.setParam("use_sim_time", false);
+	if (true == localn_.param("use_sim_time", false))
+	{
+		THROW_EXCEPTION(
+			"At present, MVSIM can only work with use_sim_time=false");
+	}
 #endif
 
-	mvsim_world_.registerCallbackOnObservation(
+	mvsim_world_->registerCallbackOnObservation(
 		[this](
 			const mvsim::Simulable& veh,
 			const mrpt::obs::CObservation::Ptr& obs) {
@@ -234,7 +254,7 @@ void MVSimNode::loadWorldModel(const std::string& world_xml_file)
 
 	// Load from XML:
 	rapidxml::file<> fil_xml(world_xml_file.c_str());
-	mvsim_world_.load_from_XML(fil_xml.data(), world_xml_file);
+	mvsim_world_->load_from_XML(fil_xml.data(), world_xml_file);
 
 	ROS12_INFO("[MVSimNode] World file load done.");
 	world_init_ok_ = true;
@@ -247,12 +267,23 @@ void MVSimNode::loadWorldModel(const std::string& world_xml_file)
  * ~MVSimNode()
  * Destructor.
  *----------------------------------------------------------------------------*/
-MVSimNode::~MVSimNode()
+MVSimNode::~MVSimNode() { terminateSimulation(); }
+
+void MVSimNode::terminateSimulation()
 {
+	if (!mvsim_world_) return;
+
+	mvsim_world_->simulator_must_close(true);
+
 	thread_params_.closing = true;
 	if (thGUI_.joinable()) thGUI_.join();
 
-	mvsim_world_.free_opengl_resources();
+	mvsim_world_->free_opengl_resources();
+
+	ros_publisher_workers_.clear();
+	// Don't destroy mvsim_server_ yet, since "world" needs to unregister.
+	mvsim_world_.reset();
+	std::cout << "[MVSimNode::terminateSimulation] All done." << std::endl;
 }
 
 #if PACKAGE_ROS_VERSION == 1
@@ -268,8 +299,8 @@ void MVSimNode::configCallback(
 	//  message = config.message.c_str();
 	ROS_INFO("MVSimNode::configCallback() called.");
 
-	if (mvsim_world_.is_GUI_open() && !config.show_gui)
-		mvsim_world_.close_GUI();
+	if (mvsim_world_->is_GUI_open() && !config.show_gui)
+		mvsim_world_->close_GUI();
 }
 #endif
 
@@ -281,6 +312,8 @@ void MVSimNode::spin()
 	using mrpt::DEG2RAD;
 	using mrpt::RAD2DEG;
 
+	if (!mvsim_world_) return;
+
 	// Do simulation itself:
 	// ========================================================================
 	// Handle 1st iter:
@@ -289,18 +322,16 @@ void MVSimNode::spin()
 	double t_new = realtime_tictac_.Tac();
 	double incr_time = realtime_factor_ * (t_new - t_old_);
 
-	if (incr_time < mvsim_world_.get_simul_timestep())	// Just in case the
-														// computer is *really
-														// fast*...
-		return;
+	// Just in case the computer is *really fast*...
+	if (incr_time < mvsim_world_->get_simul_timestep()) return;
 
 	// Simulate:
-	mvsim_world_.run_simulation(incr_time);
+	mvsim_world_->run_simulation(incr_time);
 
 	// t_old_simul = world.get_simul_time();
 	t_old_ = t_new;
 
-	const auto& vehs = mvsim_world_.getListOfVehicles();
+	const auto& vehs = mvsim_world_->getListOfVehicles();
 
 	// Publish new state to ROS
 	// ========================================================================
@@ -357,6 +388,7 @@ void MVSimNode::spin()
 					ControllerBaseInterface::TeleopInput teleop_in;
 					ControllerBaseInterface::TeleopOutput teleop_out;
 					teleop_in.keycode = keyevent.keycode;
+					teleop_in.js = mvsim_world_->getJoystickState();
 					controller->teleop_interface(teleop_in, teleop_out);
 					txt2gui_tmp += teleop_out.append_gui_lines;
 				}
@@ -389,7 +421,7 @@ void MVSimNode::thread_update_GUI(TThreadParams& thread_params)
 				World::TUpdateGUIParams guiparams;
 				guiparams.msg_lines = obj->msg2gui_;
 
-				obj->mvsim_world_.update_GUI(&guiparams);
+				obj->mvsim_world_->update_GUI(&guiparams);
 
 				// Send key-strokes to the main thread:
 				if (guiparams.keyevent.keycode != 0)
@@ -400,11 +432,11 @@ void MVSimNode::thread_update_GUI(TThreadParams& thread_params)
 			}
 			else if (obj->world_init_ok_ && obj->headless_)
 			{
-				obj->mvsim_world_.internalGraphicsLoopTasksForSimulation();
+				obj->mvsim_world_->internalGraphicsLoopTasksForSimulation();
 
 				std::this_thread::sleep_for(
 					std::chrono::microseconds(static_cast<size_t>(
-						obj->mvsim_world_.get_simul_timestep() * 1000000)));
+						obj->mvsim_world_->get_simul_timestep() * 1000000)));
 			}
 			else
 			{
@@ -476,7 +508,7 @@ void MVSimNode::notifyROSWorldIsUpdated()
 	static mrpt::system::CTicTac lastMapPublished;
 	if (lastMapPublished.Tac() > 2.0)
 	{
-		mvsim_world_.runVisitorOnWorldElements(
+		mvsim_world_->runVisitorOnWorldElements(
 			[this](mvsim::WorldElementBase& obj) {
 				publishWorldElements(obj);
 			});
@@ -484,12 +516,12 @@ void MVSimNode::notifyROSWorldIsUpdated()
 	}
 #endif
 
-	mvsim_world_.runVisitorOnVehicles(
+	mvsim_world_->runVisitorOnVehicles(
 		[this](mvsim::VehicleBase& v) { publishVehicles(v); });
 
 	// Create subscribers & publishers for each vehicle's stuff:
 	// ----------------------------------------------------
-	auto& vehs = mvsim_world_.getListOfVehicles();
+	auto& vehs = mvsim_world_->getListOfVehicles();
 	pubsub_vehicles_.clear();
 	pubsub_vehicles_.resize(vehs.size());
 	size_t idx = 0;
@@ -544,26 +576,26 @@ void MVSimNode::initPubSubs(TPubSubPerVehicle& pubsubs, mvsim::VehicleBase* veh)
 
 	pubsubs.sub_cmd_vel = n_->create_subscription<geometry_msgs::msg::Twist>(
 		vehVarName("cmd_vel", *veh), 10,
-		[this, veh](const geometry_msgs::msg::Twist::SharedPtr msg) {
+		[this, veh](const geometry_msgs::msg::Twist::ConstSharedPtr& msg) {
 			return this->onROSMsgCmdVel(msg, veh);
 		});
 #endif
 
 #if PACKAGE_ROS_VERSION == 1
 	// pub: <VEH>/odom
-	pubsubs.pub_odom =
-		n_.advertise<nav_msgs::Odometry>(vehVarName("odom", *veh), 10);
+	pubsubs.pub_odom = n_.advertise<nav_msgs::Odometry>(
+		vehVarName("odom", *veh), publisher_history_len_);
 
 	// pub: <VEH>/base_pose_ground_truth
 	pubsubs.pub_ground_truth = n_.advertise<nav_msgs::Odometry>(
-		vehVarName("base_pose_ground_truth", *veh), 10);
+		vehVarName("base_pose_ground_truth", *veh), publisher_history_len_);
 #else
 	// pub: <VEH>/odom
 	pubsubs.pub_odom = n_->create_publisher<nav_msgs::msg::Odometry>(
-		vehVarName("odom", *veh), 10);
+		vehVarName("odom", *veh), publisher_history_len_);
 	// pub: <VEH>/base_pose_ground_truth
 	pubsubs.pub_ground_truth = n_->create_publisher<nav_msgs::msg::Odometry>(
-		vehVarName("base_pose_ground_truth", *veh), 10);
+		vehVarName("base_pose_ground_truth", *veh), publisher_history_len_);
 #endif
 
 	// pub: <VEH>/chassis_markers
@@ -649,9 +681,10 @@ void MVSimNode::initPubSubs(TPubSubPerVehicle& pubsubs, mvsim::VehicleBase* veh)
 			wheel_shape_msg.points[3].z = 0;
 			wheel_shape_msg.points[4] = wheel_shape_msg.points[0];
 
-			wheel_shape_msg.color.r = w.color.R / 255.0;
-			wheel_shape_msg.color.g = w.color.G / 255.0;
-			wheel_shape_msg.color.b = w.color.B / 255.0;
+			wheel_shape_msg.color.r = w.color.R / 255.0f;
+			wheel_shape_msg.color.g = w.color.G / 255.0f;
+			wheel_shape_msg.color.b = w.color.B / 255.0f;
+			wheel_shape_msg.color.a = 1.0f;
 
 			// Set local pose of the wheel wrt the vehicle:
 			wheel_shape_msg.pose = mrpt2ros::toROS_Pose(w.pose());
@@ -731,7 +764,7 @@ void MVSimNode::onROSMsgCmdVel(
 #if PACKAGE_ROS_VERSION == 1
 	const geometry_msgs::Twist::ConstPtr& cmd,
 #else
-	const geometry_msgs::msg::Twist::SharedPtr cmd,
+	const geometry_msgs::msg::Twist::ConstSharedPtr& cmd,
 #endif
 	mvsim::VehicleBase* veh)
 {
@@ -759,23 +792,32 @@ void MVSimNode::onROSMsgCmdVel(
 /** Publish everything to be published at each simulation iteration */
 void MVSimNode::spinNotifyROS()
 {
-	using namespace mvsim;
-	const auto& vehs = mvsim_world_.getListOfVehicles();
+	if (!mvsim_world_) return;
 
-	// Get current simulation time (for messages) and publish "/clock"
-	// ----------------------------------------------------------------
+	using namespace mvsim;
+	const auto& vehs = mvsim_world_->getListOfVehicles();
+
+	// skip if the node is already shutting down:
 #if PACKAGE_ROS_VERSION == 1
-	// sim_time_.fromSec(mvsim_world_.get_simul_time());
-	// clockMsg_.clock = sim_time_;
-	// pub_clock_.publish(clockMsg_);
+	if (!ros::ok()) return;
 #else
-	// sim_time_ = n_->get_clock()->now();
-	// MRPT_TODO("Publish /clock for ROS2 too?");
+	if (!rclcpp::ok()) return;
+#endif
+
+		// Get current simulation time (for messages) and publish "/clock"
+		// ----------------------------------------------------------------
+#if PACKAGE_ROS_VERSION == 1
+		// sim_time_.fromSec(mvsim_world_->get_simul_time());
+		// clockMsg_.clock = sim_time_;
+		// pub_clock_.publish(clockMsg_);
+#else
+		// sim_time_ = n_->get_clock()->now();
+		// MRPT_TODO("Publish /clock for ROS2 too?");
 #endif
 
 #if PACKAGE_ROS_VERSION == 2
 	// In ROS2,latching doesn't work, we must re-publish on a regular basis...
-	mvsim_world_.runVisitorOnWorldElements(
+	mvsim_world_->runVisitorOnWorldElements(
 		[this](mvsim::WorldElementBase& obj) { publishWorldElements(obj); });
 #endif
 
@@ -862,10 +904,6 @@ void MVSimNode::spinNotifyROS()
 
 					// TF: /map -> <Ri>/odom
 					{
-						MRPT_TODO(
-							"Save initial pose for each vehicle, set odometry "
-							"from that pose");
-
 						Msg_TransformStamped tx;
 						tx.header.frame_id = "map";
 						tx.child_frame_id = sOdomName;
@@ -959,6 +997,13 @@ void MVSimNode::onNewObservation(
 
 	using mrpt::obs::CObservation2DRangeScan;
 
+	// skip if the node is already shutting down:
+#if PACKAGE_ROS_VERSION == 1
+	if (!ros::ok()) return;
+#else
+	if (!rclcpp::ok()) return;
+#endif
+
 	ASSERT_(obs);
 	ASSERT_(!obs->sensorLabel.empty());
 
@@ -995,6 +1040,12 @@ void MVSimNode::onNewObservation(
 	{
 		internalOn(veh, *oPC);
 	}
+	else if (const auto* oIMU =
+				 dynamic_cast<const mrpt::obs::CObservationIMU*>(obs.get());
+			 oIMU)
+	{
+		internalOn(veh, *oIMU);
+	}
 	else
 	{
 		// Don't know how to emit this observation to ROS!
@@ -1021,7 +1072,7 @@ std::string MVSimNode::vehVarName(
 {
 	using namespace std::string_literals;
 
-	if (mvsim_world_.getListOfVehicles().size() == 1)
+	if (mvsim_world_->getListOfVehicles().size() == 1)
 	{
 		return sVarName;
 	}
@@ -1035,6 +1086,7 @@ void MVSimNode::internalOn(
 	const mvsim::VehicleBase& veh,
 	const mrpt::obs::CObservation2DRangeScan& obs)
 {
+	auto lck = mrpt::lockHelper(pubsub_vehicles_mtx_);
 	TPubSubPerVehicle& pubs = pubsub_vehicles_[veh.getVehicleIndex()];
 
 	// Create the publisher the first time an observation arrives:
@@ -1046,12 +1098,13 @@ void MVSimNode::internalOn(
 	{
 #if PACKAGE_ROS_VERSION == 1
 		pub = n_.advertise<sensor_msgs::LaserScan>(
-			vehVarName(obs.sensorLabel, veh), 10);
+			vehVarName(obs.sensorLabel, veh), publisher_history_len_);
 #else
 		pub = n_->create_publisher<sensor_msgs::msg::LaserScan>(
-			vehVarName(obs.sensorLabel, veh), 10);
+			vehVarName(obs.sensorLabel, veh), publisher_history_len_);
 #endif
 	}
+	lck.unlock();
 
 #if PACKAGE_ROS_VERSION == 2
 	rclcpp::Publisher<sensor_msgs::msg::LaserScan>::SharedPtr pubLidar =
@@ -1099,9 +1152,81 @@ void MVSimNode::internalOn(
 	}
 }
 
+void MVSimNode::internalOn(
+	const mvsim::VehicleBase& veh, const mrpt::obs::CObservationIMU& obs)
+{
+	auto lck = mrpt::lockHelper(pubsub_vehicles_mtx_);
+	TPubSubPerVehicle& pubs = pubsub_vehicles_[veh.getVehicleIndex()];
+
+	// Create the publisher the first time an observation arrives:
+	const bool is_1st_pub =
+		pubs.pub_sensors.find(obs.sensorLabel) == pubs.pub_sensors.end();
+	auto& pub = pubs.pub_sensors[obs.sensorLabel];
+
+	if (is_1st_pub)
+	{
+#if PACKAGE_ROS_VERSION == 1
+		pub = n_.advertise<sensor_msgs::Imu>(
+			vehVarName(obs.sensorLabel, veh), publisher_history_len_);
+#else
+		pub = n_->create_publisher<sensor_msgs::msg::Imu>(
+			vehVarName(obs.sensorLabel, veh), publisher_history_len_);
+		ASSERT_(pub);
+#endif
+	}
+	lck.unlock();
+
+#if PACKAGE_ROS_VERSION == 2
+	rclcpp::Publisher<sensor_msgs::msg::Imu>::SharedPtr pubImu =
+		std::dynamic_pointer_cast<rclcpp::Publisher<sensor_msgs::msg::Imu>>(
+			pub);
+	ASSERT_(pubImu);
+#endif
+
+	const std::string sSensorFrameId = vehVarName(obs.sensorLabel, veh);
+
+	// Send TF:
+	mrpt::poses::CPose3D sensorPose;
+	obs.getSensorPose(sensorPose);
+
+	tf2::Transform transform = mrpt2ros::toROS_tfTransform(sensorPose);
+
+	Msg_TransformStamped tfStmp;
+	tfStmp.transform = tf2::toMsg(transform);
+	tfStmp.child_frame_id = sSensorFrameId;
+	tfStmp.header.frame_id = vehVarName("base_link", veh);
+	tfStmp.header.stamp = myNow();
+	tf_br_.sendTransform(tfStmp);
+
+	// Send observation:
+	{
+		// Convert observation MRPT -> ROS
+#if PACKAGE_ROS_VERSION == 1
+		sensor_msgs::Imu msg_imu;
+		geometry_msgs::Pose msg_pose_imu;
+		std_msgs::Header msg_header;
+#else
+		sensor_msgs::msg::Imu msg_imu;
+		std_msgs::msg::Header msg_header;
+#endif
+		// Force usage of simulation time:
+		msg_header.stamp = myNow();
+		msg_header.frame_id = sSensorFrameId;
+
+		mrpt2ros::toROS(obs, msg_header, msg_imu);
+
+#if PACKAGE_ROS_VERSION == 1
+		pub.publish(msg_imu);
+#else
+		pubImu->publish(msg_imu);
+#endif
+	}
+}
+
 void MVSimNode::internalOn(
 	const mvsim::VehicleBase& veh, const mrpt::obs::CObservationImage& obs)
 {
+	auto lck = mrpt::lockHelper(pubsub_vehicles_mtx_);
 	TPubSubPerVehicle& pubs = pubsub_vehicles_[veh.getVehicleIndex()];
 
 	// Create the publisher the first time an observation arrives:
@@ -1113,12 +1238,13 @@ void MVSimNode::internalOn(
 	{
 #if PACKAGE_ROS_VERSION == 1
 		pub = n_.advertise<sensor_msgs::Image>(
-			vehVarName(obs.sensorLabel, veh), 10);
+			vehVarName(obs.sensorLabel, veh), publisher_history_len_);
 #else
 		pub = n_->create_publisher<sensor_msgs::msg::Image>(
-			vehVarName(obs.sensorLabel, veh), 10);
+			vehVarName(obs.sensorLabel, veh), publisher_history_len_);
 #endif
 	}
+	lck.unlock();
 
 #if PACKAGE_ROS_VERSION == 2
 	rclcpp::Publisher<sensor_msgs::msg::Image>::SharedPtr pubImg =
@@ -1171,6 +1297,7 @@ void MVSimNode::internalOn(
 {
 	using namespace std::string_literals;
 
+	auto lck = mrpt::lockHelper(pubsub_vehicles_mtx_);
 	TPubSubPerVehicle& pubs = pubsub_vehicles_[veh.getVehicleIndex()];
 
 	const auto lbPoints = obs.sensorLabel + "_points"s;
@@ -1187,15 +1314,17 @@ void MVSimNode::internalOn(
 	{
 #if PACKAGE_ROS_VERSION == 1
 		pubPts = n_.advertise<sensor_msgs::PointCloud2>(
-			vehVarName(lbPoints, veh), 10);
-		pubImg = n_.advertise<sensor_msgs::Image>(vehVarName(lbImage, veh), 10);
+			vehVarName(lbPoints, veh), publisher_history_len_);
+		pubImg = n_.advertise<sensor_msgs::Image>(
+			vehVarName(lbImage, veh), publisher_history_len_);
 #else
 		pubImg = n_->create_publisher<sensor_msgs::msg::Image>(
-			vehVarName(lbImage, veh), 10);
+			vehVarName(lbImage, veh), publisher_history_len_);
 		pubPts = n_->create_publisher<sensor_msgs::msg::PointCloud2>(
-			vehVarName(lbPoints, veh), 10);
+			vehVarName(lbPoints, veh), publisher_history_len_);
 #endif
 	}
+	lck.unlock();
 
 #if PACKAGE_ROS_VERSION == 2
 	rclcpp::Publisher<sensor_msgs::msg::PointCloud2>::SharedPtr pubPoints =
@@ -1306,6 +1435,7 @@ void MVSimNode::internalOn(
 {
 	using namespace std::string_literals;
 
+	auto lck = mrpt::lockHelper(pubsub_vehicles_mtx_);
 	TPubSubPerVehicle& pubs = pubsub_vehicles_[veh.getVehicleIndex()];
 
 	const auto lbPoints = obs.sensorLabel + "_points"s;
@@ -1320,12 +1450,13 @@ void MVSimNode::internalOn(
 	{
 #if PACKAGE_ROS_VERSION == 1
 		pubPts = n_.advertise<sensor_msgs::PointCloud2>(
-			vehVarName(lbPoints, veh), 10);
+			vehVarName(lbPoints, veh), publisher_history_len_);
 #else
 		pubPts = n_->create_publisher<sensor_msgs::msg::PointCloud2>(
-			vehVarName(lbPoints, veh), 10);
+			vehVarName(lbPoints, veh), publisher_history_len_);
 #endif
 	}
+	lck.unlock();
 
 #if PACKAGE_ROS_VERSION == 2
 	rclcpp::Publisher<sensor_msgs::msg::PointCloud2>::SharedPtr pubPoints =
@@ -1393,4 +1524,4 @@ void MVSimNode::internalOn(
 		pubPoints->publish(msg_pts);
 #endif
 	}
-}
\ No newline at end of file
+}
